import { type VAny } from "../lib/core.js";
/**
 * Creates template with HTML Elements.
 *
 * ### Element Syntax:
 *
 * - `<div name="value" />` - Static Attribute.
 * - `<div name />` - Static Attribute.
 * - `<div name=${expr} />` - Dynamic attribute `element.setAttribute(name, expr)`.
 * - `<div .name=${expr} />` - Property `element[name] = expr`.
 * - `<div *name=${expr} />` - Property `element[name] = expr`, diffs against a DOM value.
 * - `<div ~name="value" />` - Static style `<div style="name:value;">`
 * - `<div ~name=${expr} />` - Dynamic style `element.style.setProperty(name, expr)`
 * - `<div @name=${expr} />` - Event `element.addEventListener(name, expr)`
 * - `<div ${directive} />` - Client-Side Directive `directive(element)`
 *
 * @example
 *
 *     function MyApp(content) {
 *       return html`
 *         <div id="App">
 *           <h1 class="Title">MyApp</h1>
 *           <div>Content ${content}</div>
 *         </div>
 *       `;
 *     }
 */
export declare const html: (strings: TemplateStringsArray, ...exprs: any[]) => VAny;
/**
 * Creates template with SVG Elements.
 *
 * ### Element Syntax:
 *
 * - `<div name="value" />` - Static Attribute.
 * - `<div name />` - Static Attribute.
 * - `<div name=${expr} />` - Dynamic attribute `element.setAttribute(name, expr)`.
 * - `<div .name=${expr} />` - Property `element[name] = expr`.
 * - `<div *name=${expr} />` - Property `element[name] = expr`, diffs against a DOM value.
 * - `<div ~name="value" />` - Static style `<div style="name:value;">`
 * - `<div ~name=${expr} />` - Dynamic style `element.style.setProperty(name, expr)`
 * - `<div @name=${expr} />` - Event `element.addEventListener(name, expr)`
 * - `<div ${directive} />` - Client-Side Directive `directive(element)`
 *
 * @example
 *
 *     function Star(content) {
 *       return svg`
 *         <svg height="210" width="500">
 *           <polygon
 *             points="100,10 40,198 190,78 10,78 160,198"
 *             ~fill="lime"
 *             ~stroke="purple"
 *             ~stroke-width="5"
 *             ~fill-rule="nonzero"
 *           />
 *         </svg>
 *       `;
 *     }
 */
export declare const svg: (strings: TemplateStringsArray, ...exprs: any[]) => VAny;
//# sourceMappingURL=index.d.ts.map