export { isVoidElement } from "./shared.js";
export declare class TemplateParserError extends Error {
    staticsOffset: number;
    textOffset: number;
    constructor(message: string, staticsOffset: number, textOffset: number);
}
export declare class TemplateScanner {
    readonly statics: string[] | TemplateStringsArray;
    readonly exprCount: number;
    text: string;
    i: number;
    e: number;
    constructor(statics: string[] | TemplateStringsArray);
    isEnd(): boolean;
    peekCharCode(i?: number): number;
    charCode(c: number): boolean;
    peekExpr(): number;
    expr(): number;
    peekString(s: string): boolean;
    string(s: string): boolean;
    peekRegExp(re: RegExp): string | undefined;
    regExp(re: RegExp): string | undefined;
}
/**
 * Formats compiler errors.
 *
 * @param statics Statics.
 * @param errorMsg Error message.
 * @param staticsOffset Expression offset.
 * @param textOffset Text offset.
 * @returns Formatted error message.
 */
export declare const formatError: (statics: TemplateStringsArray, errorMsg: string, staticsOffset: number, textOffset: number) => string;
/**
 * ASCII Char Codes.
 */
export declare const enum CharCode {
    /** "\\t" */ Tab = 9,
    /** "\\n" */ Newline = 10,
    /** "\\v" */ VerticalTab = 11,
    /** "\\r" */ CarriageReturn = 13,
    /** [space] */ Space = 32,
    /** "!" */ ExclamationMark = 33,
    /** "\\"" */ DoubleQuote = 34,
    /** "#" */ NumberSign = 35,
    /** "#" */ Hash = 35,
    /** "$" */ DollarSign = 36,
    /** "%" */ PercentSign = 37,
    /** "&" */ Ampersand = 38,
    /** "'" */ SingleQuote = 39,
    /** "(" */ LeftParen = 40,
    /** ")" */ RightParen = 41,
    /** "*" */ Asterisk = 42,
    /** "+" */ PlusSign = 43,
    /** "," */ Comma = 44,
    /** "-" */ MinusSign = 45,
    /** "." */ Dot = 46,
    /** "/" */ Slash = 47,
    /** ":" */ Colon = 58,
    /** ";" */ Semicolon = 59,
    /** "<" */ LessThan = 60,
    /** "=" */ EqualsTo = 61,
    /** ">" */ MoreThan = 62,
    /** "?" */ QuestionMark = 63,
    /** "@" */ AtSign = 64,
    /** "[" */ LeftSquareBracket = 91,
    /** "\" */ Backslash = 92,
    /** "]" */ RightSquareBracket = 93,
    /** "^" */ Caret = 94,
    /** "_" */ Underscore = 95,
    /** "`" */ GraveAccent = 96,
    /** "{" */ LeftCurlyBrace = 123,
    /** "|" */ VerticalBar = 124,
    /** "}" */ RightCurlyBrace = 125,
    /** "~" */ Tilde = 126
}
//# sourceMappingURL=parser.d.ts.map